/*******************************************************************************
  Filename:       npi.c
  Revised:        $Date: 2008-06-11 14:30:47 -0700 (Wed, 11 Jun 2008) $
  Revision:       $Revision: 17210 $

  Description:    This file contains the Network Processor Interface (NPI),
                  which abstracts the physical link between the Application
                  Processor (AP) and the Network Processor (NP). The NPI
                  serves as the HAL's client for the SPI and UART drivers, and
                  provides API and callback services for its client.

  Copyright 2006-2014 Texas Instruments Incorporated. All rights reserved.

  IMPORTANT: Your use of this Software is limited to those specific rights
  granted under the terms of a software license agreement between the user
  who downloaded the software, his/her employer (which must be your employer)
  and Texas Instruments Incorporated (the "License").  You may not use this
  Software unless you agree to abide by the terms of the License. The License
  limits your use, and you acknowledge, that the Software may not be modified,
  copied or distributed unless embedded on a Texas Instruments microcontroller
  or used solely and exclusively in conjunction with a Texas Instruments radio
  frequency transceiver, which is integrated into your product.  Other than for
  the foregoing purpose, you may not use, reproduce, copy, prepare derivative
  works of, modify, distribute, perform, display or sell this Software and/or
  its documentation for any purpose.

  YOU FURTHER ACKNOWLEDGE AND AGREE THAT THE SOFTWARE AND DOCUMENTATION ARE
  PROVIDED AS IS WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
  INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY, TITLE,
  NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL
  TEXAS INSTRUMENTS OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER CONTRACT,
  NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR OTHER
  LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
  INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE
  OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT
  OF SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
  (INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.

  Should you have any questions regarding your right to use this Software,
  contact Texas Instruments Incorporated at www.TI.com.
*******************************************************************************/

/*******************************************************************************
 * INCLUDES
 */
#include <string.h>
#include <xdc/std.h>
#include <ti/sysbios/family/arm/cc26xx/Power.h>
#include <ti/sysbios/family/arm/cc26xx/PowerCC2650.h>
#include <ti/sysbios/family/arm/m3/Hwi.h>
#include <ti/drivers/pin/PINCC26XX.h>
#include "inc/hw_memmap.h"
#include "inc/hw_ints.h"
#include "Board.h"
#include "hal_types.h"
#include "npi.h"
#include "npi_config.h"
#include "hci_tl.h"
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Swi.h>
#include <ti/drivers/UART.h>
#include <ti/drivers/uart/UARTCC26XX.h>
#include "simpleBLEPeripheral.h"
/*******************************************************************************
 * MACROS
 */

/*******************************************************************************
 * CONSTANTS
 */
// Hardware Failure Status (originally defined in ll_common.h)
#define HW_FAIL_OUT_OF_MEMORY     0x81

/*******************************************************************************
 * LOCAL VARIABLES
 */
static volatile bool npiPMSetConstraint = false;  // flag for low power mode
static volatile bool npiRxActive        = false;  // flag for ongoing NPI RX
static volatile bool npiTxActive        = false;  // flag for ongoing NPI TX

// App/NPI buffers
static uint16 npiRxIdxTail   = 0;
static uint16 npiRxIdxHead   = 0;
static Char npiRxBuf[NPI_UART_RX_BUF_SIZE];       // NPI Receive buffer
static uint16_t npiRxCount = 0;                   // NPI Receive buffer count

static UART_Handle uartHandle;
static Char isrRxBuf[UART_ISR_BUF_SIZE];         // UART ISR RX buffer

//RTOS callback for TX/RX
npiRtosCB_t appTxCB;
npiRtosCB_t appRxCB;

PIN_Config npiHandshakePinsCfg[] =
{
    RTS_PIN   | PIN_GPIO_OUTPUT_DIS  | PIN_INPUT_EN  |  PIN_PULLUP,
    CTS_PIN   | PIN_GPIO_OUTPUT_EN   | PIN_GPIO_HIGH |  PIN_PUSHPULL | PIN_DRVSTR_MAX,
    PIN_TERMINATE
};

PIN_State  npiHandshakePins;
PIN_Handle hNpiHandshakePins;

/*******************************************************************************
 * GLOBAL VARIABLES
 */

/*******************************************************************************
 * EXTERNAL VARIABLES
 */
extern UARTCC26XX_Object uartCC26XXObjects[];

/*******************************************************************************
 * PROTOTYPES
 */

/*******************************************************************************
 * LOCAL FUNCTIONS
 */
static uint16 NPI_readIsrBuf(int size);
static void NPI_writeCallBack (UART_Handle handle, void *ptr, size_t size);
static void NPI_readCallBack (UART_Handle handle, void *ptr, size_t size);
#if defined(NPI_UART_FC) && NPI_UART_FC==TRUE
static void RTSPinHwiFxn(PIN_Handle hPin, PIN_Id pinId);
#endif
/*******************************************************************************
 * @fn          NPI_InitTransport
 *
 * @brief       This routine initializes the transport layer and opens the port
 *              of the device. Note that based on project defines, either the
 *              UART, USB (CDC), or SPI driver can be used.
 *
 * input parameters
 *
 * @param      None
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      None.
 */
void NPI_InitTransport(npiCBack_t npiCBack)
{  
  return;
}

/*******************************************************************************
 * @fn          NPI_getTxState
 *
 * @brief       This routine returns the state of transmission on NPI
 *
 * input parameters
 *
 * @param      None
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      state of NPI transmission - 1 - active, 0 - not active
 */
bool inline NPI_getTxState(void)
{  
  return npiTxActive;
}

/*******************************************************************************
 * @fn          NPI_getRxState
 *
 * @brief       This routine returns the state of reception on NPI
 *
 * input parameters
 *
 * @param      None
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      state of NPI transmission - 1 - active, 0 - not active
 */
bool inline NPI_getRxState(void)
{  
#ifdef POWER_SAVING
  return npiRxActive;
#else
  return false;
#endif
}

/*******************************************************************************
 * @fn          NPI_InitInterface
 *
 * @brief       This routine can open a serial interface port and assign
 *              read and write callbacks. This is done under the assumption
 *              that read and write functionality is implemented with callbacks
 *
 * input parameters
 *
 * @param      port - Port number unique to serial interface
 * @param      npiCBTx - function callback pointer for write
 * @param      npiCBRx - function callback pointer for read
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      None.
 */
extern SYSTEMPARAMETER SystemParameter; 
void NPI_InitInterface(uint8_t interfaceType, npiRtosCB_t npiCBTx,
                       npiRtosCB_t npiCBRx)
{
  if (interfaceType == UART_INTERFACE)
  {
    UART_Params params;

    // Configure UART parameters.
    UART_Params_init(&params);
    if(SystemParameter.UartBPS==0)
        params.baudRate      = NPI_UART_BR_115200;
	else if(SystemParameter.UartBPS==1)
		params.baudRate      = NPI_UART_BR_57600;
	else if(SystemParameter.UartBPS==2)
		params.baudRate      = NPI_UART_BR_38400;
	else if(SystemParameter.UartBPS==3)
		params.baudRate      = NPI_UART_BR_19200;
	else if(SystemParameter.UartBPS==4)
		params.baudRate      = NPI_UART_BR_9600;
	else if(SystemParameter.UartBPS==5)
		params.baudRate      = NPI_UART_BR_4800;
    params.readDataMode  = UART_DATA_BINARY;
    params.writeDataMode = UART_DATA_BINARY;
    params.dataLength    = UART_LEN_8;
    params.stopBits      = UART_STOP_ONE;
    params.readMode      = UART_MODE_CALLBACK;
    params.writeMode     = UART_MODE_CALLBACK;
    params.readEcho      = UART_ECHO_OFF;
    params.readTimeout   = 5;

    params.readCallback  = NPI_readCallBack;
    params.writeCallback = NPI_writeCallBack;

    // Application callbacks are nested within NPI callbacks
    appTxCB = npiCBTx;
    appRxCB = npiCBRx;

    // Open / power on the UART.
    uartHandle = UART_open(Board_UART, &params);
    //Enable Partial Reads on all subsequent UART_read()ʹֶȡк
    UART_control(uartHandle, UARTCC26XX_RETURN_PARTIAL_ENABLE,  NULL); 
    
    // This call will set the read size for ensuring HWI callback
    NPI_uartRead();

#if defined(NPI_UART_FC) && NPI_UART_FC==TRUE
    CTS_DISABLE();

    // Initialize SRDY/MRDY. Enable int after callback registered
    hNpiHandshakePins = PIN_open(&npiHandshakePins, npiHandshakePinsCfg);
    PIN_registerIntCb(hNpiHandshakePins, RTSPinHwiFxn);
    PIN_setConfig(hNpiHandshakePins, PIN_BM_IRQ,          RTS_PIN | PIN_IRQ_BOTHEDGES);
    
    // Enable wakeup
    PIN_setConfig(hNpiHandshakePins, PINCC26XX_BM_WAKEUP, RTS_PIN | PINCC26XX_WAKEUP_NEGEDGE);
#endif
  }
  else if (interfaceType == SPI_INTERFACE)
  {

  }
}
void NPI_UART_CLOSE(void)
{
	UART_close(uartHandle);
}
/*******************************************************************************
 * @fn          NPI_setPowerMgrRequirements
 *
 * @brief       This routine is used to set constraints on power manager
 *
 * input parameters
 *
 * @param       None.
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      None.
 */
#if defined(NPI_UART_FC) && NPI_UART_FC==TRUE
static void NPI_setPowerMgrRequirements()
{
  if (npiPMSetConstraint == true)
  {
    return;
  }
#ifdef POWER_SAVING
  // set constraints for Standby, powerdown and idle mode
  Power_setConstraint  (Power_SB_DISALLOW);
  Power_setConstraint  (Power_SD_DISALLOW);
  Power_setConstraint  (Power_IDLE_PD_DISALLOW);
#endif
  npiPMSetConstraint = true;
}
#endif

/*******************************************************************************
 * @fn          NPI_releasePowerMgrRequirements
 *
 * @brief       This routine is used to release constraints on power manager
 *
 * input parameters
 *
 * @param       None.
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      None.
 */
static void NPI_releasePowerMgrRequirements()
{

  if (npiPMSetConstraint == false)
  {
    return;
  }
#ifdef POWER_SAVING
  // release constraints for Standby, powerdown and idle mode
  Power_releaseConstraint  (Power_SB_DISALLOW);
  Power_releaseConstraint  (Power_SD_DISALLOW);
  Power_releaseConstraint  (Power_IDLE_PD_DISALLOW);
#endif
  npiPMSetConstraint = false;
}


#if defined(NPI_UART_FC) && NPI_UART_FC==TRUE
static void RTSPinHwiFxn(PIN_Handle hPin, PIN_Id pinId)
{
  if (PIN_getInputValue(RTS_PIN) == 0)
  {
    // set constraints for Standby, powerdown and idle mode
    NPI_setPowerMgrRequirements(); 

    if(!npiTxActive)
    {
      npiRxActive = true;
    }

    CTS_ENABLE();
  }
  else
  {
    CTS_DISABLE();
    npiRxActive = false;

    UARTCC26XX_HWAttrs const  *hwAttrs = uartHandle->hwAttrs;

    if(UARTCharsAvail(hwAttrs->baseAddr) == false)
    {
      // release constraints for Standby, powerdown and idle mode
      NPI_releasePowerMgrRequirements();
    }

    if (UARTBusy(hwAttrs->baseAddr))
    {
      npiTxActive = true;

      CTS_ENABLE();

      NPI_setPowerMgrRequirements();
    }
  }
}
#endif

/*******************************************************************************
 * @fn          NPI_writeCallBack
 *
 * @brief       This callback is invoked on Write completion
 *
 * input parameters
 *
 * @param      handle - handle to the UART port
 * @param      ptr    - pointer to data to be transmitted
 * @param      size   - size of the data
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      None.
 */
static void NPI_writeCallBack (UART_Handle handle, void *ptr, size_t size)
{
  CTS_DISABLE();

  if (PIN_getInputValue(RTS_PIN) == 1)
  {
    // release constraints for Standby, powerdown and idle mode
    NPI_releasePowerMgrRequirements();
  }
  else
  {
    CTS_ENABLE();
  }

  npiTxActive = false;
  
  //Invoke application callback
  appTxCB(size);
}

/*******************************************************************************
 * @fn          NPI_readCallBack
 *
 * @brief       This callback is invoked on Read completion of readSize/receive
 *              timeout
 *
 * input parameters
 *
 * @param      handle - handle to the UART port
 * @param      ptr    - pointer to data to be transmitted
 * @param      size   - size of the data
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      None.
 */
static void NPI_readCallBack (UART_Handle handle, void *ptr, size_t size)
{
  UInt hwiKey = Hwi_disable();
  UInt taskKey = Task_disable();

  if ( size )
  {
    // Read the ISR buffer into the NPI buffer within the callback
    NPI_readIsrBuf(size);
    appRxCB(0);
  }
  //appRxCB(0xFF);
  UART_read(uartHandle, &isrRxBuf[0], UART_ISR_BUF_SIZE - 2);

  Task_restore(taskKey);
  Hwi_restore(hwiKey);
}

/*******************************************************************************
 * @fn          NPI_readIsrBuf
 *
 * @brief       This routine reads data from the transport layer based on len,
 *              and places it into the buffer.
 *
 * input parameters
 *
 * @param       buf - Pointer to buffer to place read data.
 * @param       len - Number of bytes to read.
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      Returns the number of bytes read from transport.
 */
static uint16 NPI_readIsrBuf(int size)
{
  uint8_t count = size;
  uint8_t idx;
  // Copy the UART buffer to the application buffer
  for (idx = 0; idx < count; idx++)
  {
    npiRxBuf[npiRxIdxTail++] = isrRxBuf[idx];
    npiRxIdxTail %= NPI_UART_RX_BUF_SIZE;
  }

  return count;
}



/*******************************************************************************
 * @fn          NPI_uartRead
 *
 * @brief       This routine reads data/activates a read from the UART buffer
 *
 * input parameters
 *
 * @param       None.
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      None.
 */
void inline NPI_uartRead()
{
//  UARTCC26XX_RETURN_PARTIAL
  UART_read(uartHandle, &isrRxBuf[0], UART_ISR_BUF_SIZE - 2);
}

/*******************************************************************************
 * @fn          NPI_ReadTransport
 *
 * @brief       This routine reads data from the transport layer based on len,
 *              and places it into the buffer.
 *
 * input parameters
 *
 * @param       buf - Pointer to buffer to place read data.
 * @param       len - Number of bytes to read.
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      Returns the number of bytes read from transport.
 */
uint16 NPI_ReadTransport(uint8 *buf, uint16 len)
{
  uint16_t idx = 0;

  for (idx = 0; idx < len; idx++)
  {
    *buf++ = npiRxBuf[npiRxIdxHead++];

    npiRxIdxHead %= NPI_UART_RX_BUF_SIZE;
  }

  return len;
}


/*******************************************************************************
 * @fn          NPI_WriteTransport
 *
 * @brief       This routine writes data from the buffer to the transport layer.
 *
 * input parameters
 *
 * @param       buf - Pointer to buffer to write data from.
 * @param       len - Number of bytes to write.
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      Returns the number of bytes written to transport.
 */
uint16 NPI_WriteTransport(uint8 *buf, uint16 len)
{
#if TI_DRIVERS_UART_INCLUDED
  CTS_ENABLE();
  npiTxActive = true;
  UART_write(uartHandle, (Char *)buf, len);
  return len;
#else
  return 0;
#endif
}

/*******************************************************************************
 * @fn          NPI_RxBufLen
 *
 * @brief       This routine returns the number of bytes in the receive buffer.
 *
 * input parameters
 *
 * @param       None.
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      Returns the number of bytes in the receive buffer.
 */
uint16 NPI_RxBufLen(void)
{
  if (npiRxIdxHead < npiRxIdxTail)
  {
    npiRxCount = npiRxIdxTail - npiRxIdxHead;
  }
  else
  {
    npiRxCount = NPI_UART_RX_BUF_SIZE - (npiRxIdxHead - npiRxIdxTail);
  }

  return npiRxCount;
}

/*******************************************************************************
 * @fn          NPI_GetMaxRxBufSize
 *
 * @brief       This routine returns the max size receive buffer.
 *
 * input parameters
 *
 * @param       None.
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      Returns the max size of the receive buffer.
 */
uint16 NPI_GetMaxRxBufSize(void)
{
  return(NPI_UART_RX_BUF_SIZE);
}


/*******************************************************************************
 * @fn          NPI_GetMaxTxBufSize
 *
 * @brief       This routine returns the max size transmit buffer.
 *
 * input parameters
 *
 * @param       None.
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      Returns the max size of the transmit buffer.
 */
uint16 NPI_GetMaxTxBufSize(void)
{
  return(NPI_UART_TX_BUF_SIZE);
}

/*******************************************************************************
 * @fn          NPI_FlushRxBuf
 *
 * @brief       This routine resets the head and tail pointers for RX buffer
 *
 * input parameters
 *
 * @param       None.
 *
 * output parameters
 *
 * @param       None.
 *
 * @return      None
 */
void NPI_FlushRxBuf(void)
{
  npiRxIdxHead = npiRxIdxTail;
  npiRxCount = 0;
}
uint8 uart_buffer[202];
uint8 uart_len=0;
uint8 pBLESend=0;
void ReadyToSendFormBLE(uint8 * p, uint8 len)
{
  memset(uart_buffer,0,202);
  uart_len=0;
  memcpy(uart_buffer,p,len);
  uart_len = len ;
  pBLESend = 0 ;
}

/*******************************************************************************
 * @fn          NPI_serialPacketParser
 *
 * @brief       This routine is the NPI Transport callback function. It builds
 *              a BLE command or data packet (depending on the packet type)
 *              from bytes received by the HCI via the interface port. Bytes
 *              are consumed until either there are none left, or there aren't
 *              enough to complete a packet, in which case it is left in the
 *              driver buffer.
 *
 *              HCI command packet format:
 *              Packet Type + Command opcode + lengh + command payload
 *              | 1 octet   |      2         |   1   |      n        |
 *
 *              HCI data packet format:
 *              Packet Type +   Conn Handle  + lengh +  data payload
 *              | 1 octet   |      2         |   2   |      n      |
 *
 * input parameters  - None
 *
 * output parameters - None
 *
 * @param       None.
 *
 * @return      None.
 */
void NPI_serialPacketParser()
{
#ifdef USE_ICALL  
  static uint8  hciPktState = HCI_PARSER_STATE_PKT_TYPE;
  static uint8  pktType;
  //static uint16 param1;   // opcode for command, connection handle for data
  //static uint16 pktLen;
  uint8 *pUrxBuffer;
  //
  uint8  done = FALSE;
  uint16 numBytes;

  // get the number of available bytes to process
  numBytes = NPI_RxBufLen();

  // check if there's any serial port data to process
	while ((numBytes > 0) && (!done))
	{
		pUrxBuffer = malloc( numBytes+10 );
		// yes, so check if we are at the start of a new packet
		if (hciPktState == HCI_PARSER_STATE_PKT_TYPE)
		{
			NPI_ReadTransport(pUrxBuffer, numBytes);
			ReadyToSendFormBLE(pUrxBuffer,numBytes);
			free(pUrxBuffer); 
			NPI_ReadTransport(&pktType, 1);
			numBytes -= 1;
			switch(pktType)
			{
				case HCI_CMD_PACKET:
					hciPktState = HCI_CMD_PARSER_STATE_OPCODE;
					break;
				case HCI_ACL_DATA_PACKET:
				case HCI_SCO_DATA_PACKET:
					hciPktState = HCI_DATA_PARSER_STATE_HANDLE;
					break;
				default:
					NPI_FlushRxBuf();
					return;
			}
		}
	}
#endif
}

/*******************************************************************************
 * This function sends a Hardware Error Event to the Host.
 *
 * Public function defined in hci_c_event.h.
 */
void HCI_HardwareErrorEvent(uint8 hwErrorCode)
{
  uint8 msg[4];
  
  msg[0] = HCI_EVENT_PACKET;
  msg[1] = HCI_BLE_HARDWARE_ERROR_EVENT_CODE;
  msg[2] = 1; // data length

  // error code
  msg[3] = hwErrorCode;
      
  NPI_WriteTransport(msg, 4);

  return;
}
/*******************************************************************************
 ******************************************************************************/
